
/* $Id: AST.h 1.26.1.8.1.15.1.14.2.11 Wed, 31 Mar 1999 23:49:41 -0800 liblit $	 -*-C++-*-							*/
/* AST.h: Definitions for Abstract Syntax Trees			*/

/* Copyright (C) 1996, Paul N. Hilfinger.  All Rights Reserved.		*/

#ifndef _AST_H_
#define _AST_H_

#define __SINST__		// g++ header bug
#include <set>
#include <string>

class ostream;
class ofstream;

#include "common.h"
#include "config.h"
#include "literal.h"
#include "llist.h"
#include "local/context.h"
#include "string16.h"
#include "utils.h"
#include "StringTable.h"

class Bblock;
class Candidate;
class CatchListNode;
class CfgNode;
class ClassDecl;
class CodeContext;
class CommonSub;
class CompileUnitNode;
class CfSource;
class CtType;
class Decl;
class Defs;
class Environ;
class EnvironIter;
class ExprNode;
class FinallyNode;
class LocalVars;
class LocalEnactContext;
class ObjectNode;
class Poly;
class TreeListNode;
class TreeNode;
class TypeListNode;
class TypeNode;
class UniqueId;

template< class T > class VectIter;
template< class T > class CnstVectIter;

#include "parse.h"
#include "template.h"

class TreeNode : public Common 
{
public:
  /*****************************************************************/

  /* Accessors  */

  /* The position in the source ascribed to this node. */
  const SourcePosn &position() const { return where; }

  /* The parent of this node in the AST. NULL for the root node */
  TreeNode *parent() const { return _parent; }

  /* The Ith child (numbering from 0) of *this, 0 <= I < arity(). */
  TreeNode* child (int i) const {
    assert ((unsigned int) i < (unsigned int) arity());
    return children[i];
  }

  typedef VectIter<TreeNode*> ChildIter;
  typedef VectIter<TypeNode*> TypeIter;
  typedef CnstVectIter<const TreeNode*> ConstChildIter;
  typedef CnstVectIter<const TypeNode*> ConstTypeIter;

  /* An iterator over all children of *THIS. */
  ChildIter allChildren();
  TypeIter allTypes();
  ConstChildIter allChildren() const;
  ConstTypeIter allTypes() const;


  /* Print a representation of *this on OS, indenting each line */
  /* after the first by INDENT spaces. */
  virtual void print (ostream& os, unsigned depth = 0) const;

  // Serialize a representation of *this.  Similar to print(), but
  // intended for processing by a machine rather than by a human.
  virtual void dump (ostream &, unsigned depth = 0) const;

  /* Return true if argument appears to be identical to this. */
  /* Implementation of compare() is generated by defnodeimpls.el. */
  virtual bool compare (const TreeNode *) const;

  /* Return a string representing this node suitable for printing */
  /* in an error message */
  virtual string errorName() const;

  /* When true, indicates an optional node that is absent (that is, */
  /* that the node comes from the static constant omitted, defined */
  /* below).  */ 
  virtual bool absent() const;

  /* A TreeNode for which absent() is true.  By convention, every */
  /* derivate of TreeNode indicating a type of node that may be */
  /* missing (i.e., that represents an optional piece of syntax) */
  /* also has a static variable of this name which absent() is true.  */
  static TreeNode* const omitted;

  /* The printed name of the operator for *this. */
  virtual const char* oper_name() const;

  /* The name of operators, for operator overloading and error messages */
  /* (undefined on non-operators) */
  virtual const char *operatorName() const;

  /* The number of children this node possesses. */
  virtual int arity() const;

  /* Make a copy of this node */
  virtual TreeNode *clone() const;

  /*****************************************************************/

  /* Modifying children */

  /* Set the Ith child (numbering from 0) of *this to NEWCHILD, 0 <= */
  /* I < arity().  The dynamic type of NEWCHILD must be correct for */
  /* the type of node. */
  void child (int i, TreeNode* newChild) {
    assert ((unsigned int) i < (unsigned int) arity());
    children[i] = newChild;
    if (newChild) newChild->_parent = this;
  }

  /*****************************************************************/

  /* Static semantics */

  /* Static analysis, phase 1: build the package environment */
  virtual void packageResolution();

  /* Static analysis, phase 2: 
     resolveClass: build the class environment
     resolveInheritance: add inherited members */
  virtual void resolveClass (Decl *package, ClassDecl *cclass, Environ *fileEnv);
  virtual void resolveInheritance();

  /* Static analysis, phase 3: resolve all names and remove parsing
     ambiguities */
  class NameContext;
  virtual TreeNode* resolveName (NameContext *ctx);

  /* Static analysis, phase 4: (global) verify inheritance circularity */

  /* Static analysis, phase 5: (global) fold constants */
  virtual TreeNode *foldConstants();
  virtual void foldFields();

  /* Static analysis, phase 6: resolve fields, method overloading, and
     compute types (lazily) */
  class FieldContext;
  virtual TreeNode *resolveField(FieldContext *ctx, bool *inAssignment);

  /* Static analysis, phase 7: simplify string concatenations */
  virtual TreeNode *resolveConcat();

  /* Static analysis, phase 8: check types */
  /* The return value is true if the statement/expression may terminate
     normally (i.e. doesn't return/throw - break & continue are ignored
     here because they don't influence whether the whole body of a
     function may terminate normally) */
  class TypeContext;
  virtual bool typecheck(TypeContext *ctx);

  /* Static analysis, phase 10: rewrite some tree nodes for the back end */
  virtual TreeNode *rewrite( const FieldContext * );

  /* Static analysis, phase 11: Single inference */
  class SingleState;
  class SingleContext;
  virtual SingleState single(SingleContext *);

  /* Static analysis, phase 12: Add widening conversions discovered by
     the type checker */
  virtual void widen();

  /* Static analysis, phase 13: Locate interclass dependencies that
      affect code generation. */
  virtual void resolveRequires(ClassDecl *);

  /* Static analysis, phase 14: Introduce temporaries */
  virtual TreeNode *introduceTemporary(TreeNode *, llist<TreeNode *> *&,
				       llist<TreeNode *> *&);

  /* Static analysis, phase 15: Introduce temporaries */
  virtual llist<CfgNode *> *makeCfg();
  virtual llist<CfgNode *> *getCfgExtent();
  virtual llist<CfgNode *> *getExceptionCfg();
  virtual CfgNode *getContinueTarget();
    
  /* The type of this node (only makes sense for ExprNodes) */
  virtual TypeNode* type();

  /* The type of constants (needed before typechecking for 
     constant folding). Returns 0 for non-constants */
  virtual Common::Kind constantType();

  /* For nodes that represent field accesses (ObjectFieldAccessNode, */
  /* ThisFieldAccessNode, SuperFieldAccessNode) the type of the */
  /* object that is accessed (e.g., for a node representing FOO.BAR, */
  /* the type of FOO. */
  virtual TypeNode *accessedObjectType() const;

  /* Return true if this node is an int constant in the specified
     range. Assumes constant folding has already been done */
  virtual bool isIntConstant(int from, int to) const;

  /* Report an error if this expression cannto be assigned with 'with' */
  virtual void checkAssignable(TreeNode *with, bool allowFinalAssign);

  /* Some methods that are type-related but that also appear in
     TreeNode's (because of TreeListNode, LitNode and NameNode) */

  /* Assuming THIS and T1 represent types, true iff they represent */
  /* identical types. */
  virtual bool typeIdent(const TreeNode* T1) const;
  /* The same, but ignore type modifiers at the top-level */
  virtual bool typeIdentNM(const TreeNode* T1) const;
  /* The same, but ignore single type modifiers at the top-level (this is getting silly) */
  virtual bool typeIdentNS(const TreeNode* T1) const;
  // Return true if corresponding types in 'this' and 'T1' have the same qualifiers
  // (requires that this and T1 be TypeListNodes with the same arity)
  bool singleArgsIdent(const TreeNode* T1) const;

  /* The JVM signature corresponding to this type. */
  virtual string signature() const;

  /* Return a human readable string for this typename. This string
     should be unique (i.e. include the package) */
  virtual string typeName() const;


  // The type conversions inherent in the xxxAssignNode nodes cannot be
  // represented by CastNode's. Instead, give a op= b, two functions are provided:
  //   - widenSource: the type to which a should be widened to perform 'a op b'
  //   - castResult: the type to which 'a op b' should be cast before assignment to a
  // In both cases, NULL is returned if no conversion is necessary.
  // None of these casts require run-time checks.
  // widenSource is also the type at which 'op' is computed

  virtual TypeNode *widenSource();
  virtual TypeNode *castResult();

  /*****************************************************************/

#ifdef USE_LOCAL_INFERENCE
  
  // Local qualification inference

  void unshareTypeTree();
  virtual void unshareType();

  void inferLocalTree();
  void inferLocalTree( const LocalContext );
  virtual void inferLocalChildren( const LocalContext ) const;
  virtual void inferLocal( const LocalContext );

  virtual bool localInferable() const;
  void localConstrainCall( const TypeNode & ) const;
  virtual void localReceiveAssign( const TypeNode & ) const;

  void localEnactTree( const LocalEnactContext & );
  virtual void localEnactChildren( const LocalEnactContext & ) const;
  virtual void localEnact( const LocalEnactContext & );
  bool localEnactBits( const LocalEnactContext &, Modifiers & ) const;
  
  void localReportChange( const string &, const TypeNode & ) const;
  
#endif // USE_LOCAL_INFERENCE

  /*****************************************************************/

  // look for a suitable main() method
  virtual void lookForMain() const;

  /* Code generation */
  virtual void codeGen ();			// classes
  virtual void codeGen (CfSource &);		// methods
  virtual void codeGen (CodeContext &);		// everything else

  // selective filtering; implemented by CompileUnitNode and ClassDeclNode
  virtual bool selectedForCodeGen(int) const;
  
  // Pass that searches for potentially optimizable index expressions.
  virtual void foreachOpt0 ();
  
  // Pass that searches for defs.  Argument is where the result is accumulated.
  virtual void findDefs (Defs *);
  
  // Whether a method call has no callee visible side effects.
  virtual bool isSideEffectFree ();
  
  // Whether a method call has no callee visible side effects and
  // always returns the same value for the same arguments.
  virtual bool isPureFunction ();

  // True for ObjectFieldAccessNode's whose object is immutable
  virtual bool isFieldOfImmutable() const;
  
  // Return the cached result of defs analysis on this node.
  virtual Defs * defs ();
  
  // Possibly optimizable index expressions with respect to this loop.
  // Defined on ForEachStmtNode only.
  virtual llist<Candidate *> *& candidateIndexExprs();

  // Points to the candidate corresponding to the innermost foreach-introduced
  // variable in the index expression of the node.
  // Defined on ArrayAccessNode only.
  virtual llist<Candidate *> *       & innerForeachIterationVar();
  virtual llist<Candidate *> * const & innerForeachIterationVar() const;

  // Find loop invariants with respect to a certain loop
  virtual bool loopInvariant(TreeNode *WRTloop);

  // Lift loop invariants
  virtual void liftInvariants( CodeContext &, TreeNode *WRTloop );

  // Find expressions that "look the same" in a given subtree
  virtual void *commonSubExpr();

  // Return invariants for a loop.  Defined on ForEachStmtNode only.
  virtual llist<TreeNode *> *& invariants();

  // Searches a candidate index expr for points that might be needed
  // by *other* foreach's.
  virtual void needPointAnal(TreeNode *candidate_fp);

  // Return whether a point in a ForEachPairNode needs to be constructed.
  virtual bool& needPoint();

  // Find induction variables with respect to a certain loop
  virtual bool inductionVar(TreeNode *WRTloop);

  // Return induction vars for a loop.  Defined on ForEachStmtNode only.
  virtual llist<TreeNode *> *& inductionVars();

  // Emit debug information
  virtual TreeNode *emitDebugInfo(ofstream &os, StringTable &st);


  // implemented by StatementNode
  virtual void prepareBreak();

  // implemented by StatementNode and friends
  virtual void unwind( CodeContext & ) const;
  virtual void unwindReturn( CodeContext &, const string & ) const;
  virtual void unwindBreak( CodeContext &, const TreeNode & );
  virtual void unwindContinue( CodeContext &, const TreeNode & ) const;

  // implemented by TryStmtNode
  virtual void emitCleanup( CodeContext & ) const;
  virtual const string emitUse( const ObjectNode & ) const;
  
  // implemented by CatchNode
  virtual void emitCatchTable( ostream & ) const;

  // implemented by ExprNode
  virtual const string declareTemporary( LocalVars & );
  virtual const string defineTemporary( CodeContext & ); // also implemented by some subclasses
  /*   */ const string buildTemporary( CodeContext &, const string & );
  virtual const string simpleVar( CodeContext & );
  virtual const string emitExpression( CodeContext & );
  virtual const string emitExpression2( CodeContext & );
  virtual const string getLvalue( CodeContext & );
  virtual const bool isLocalLvalue();
  virtual const bool isSimpleLvalue();

  // overridden by ArrayAccessNode
  virtual bool isArrayAccessNode() const;

  // implemented by ExprNode subclasses
  virtual const string receiveAssign( const string &, const string &, AssignKind );

  // implemented by FieldAccessNode subclasses
  virtual const string emitMethodCall( CodeContext &, TreeListNode & );

  // whether this node is in the StatementNode hierarchy or not
  virtual bool isStatementNode(void) { return false; }

  // Polynomial analysis.
  Poly *getPoly(void) { return _poly; }
  void setPoly(Poly *poly);
  void setPoly(TreeNode *expr, bool inv, bool p_iv, bool s_iv);
  virtual void calcPoly(void) { return; }

  // Identify objects that have lvals.
  virtual bool hasLval() { return false; }

  // implemented by NameNode
  virtual void setDefs(llist<TreeNode *> *defs) {} ;
  virtual llist<TreeNode *> *getDefs(void) { return NULL; };
  virtual void setUses(llist<TreeNode *> *uses) {} ;
  virtual llist<TreeNode *> *getUses(void) { return NULL; };

  /*****************************************************************/

  /* The definitions of all accessors for all possible children and */
  /* attributes.   By default, all of these cause runtime errors */
  /* when called; they are overridden in classes that define them. */ 
  #include "tree-accessors.h"

  /* Template instantiatiation on this node, with the specified
     substitution */
  virtual TreeNode *instantiate(Subst *args);

  // resolveClass helper functions
  virtual bool updateConstructorValidity();

protected:
  /* Set line number from children, if not determined by P. */
  void setWhere (SourcePosn p);

  /* Report a fatal error and exit. */
  DOES_NOT_RETURN void fatal (const string& msg) const;

  /* Report that name is undefined and exit. */
  DOES_NOT_RETURN void undefined (const string& name) const;

public:
  // Issue an error, warning, or message at the node's source position
  void error(   const string & ) const;
  void warning( const string & ) const;
  void message( const string & ) const;

  /* Report that a feature is not yet implemented and exit. */
  DOES_NOT_RETURN void unimplemented( const string & ) const;

protected:
  TreeNode() : _parent(NULL), _poly(NULL) { }
  SourcePosn where;

  TreeNode *_parent;
  TreeNode** children;

  /* Useful error checking primitives */
  /* Report error if a variable of type to cannot be initialised
     with expression from. Returns false if error reported. Call
     operatorName() to get a name for this node. */
  void assertArithType(TypeNode *type);
  void assertBooleanType(TypeNode *type);
  void assertIntegralType(TypeNode *type);
  void assertPromoteableInt(TypeNode *type);
  void assertReferenceType(TypeNode *type);
  void assertDomainType(TypeNode *type);
  void assertRectDomainType(TypeNode *type);

  void assertAssignable(TypeContext *ctx, TreeNode *tree, TreeNode *with);
  void assertCastable(TypeNode *from, TypeNode *to);

  // typecheck helper functions
  virtual bool _typecheck(TypeContext *ctx);
  void buildAllowedExceptions(TypeContext *ctx);
  void buildAllowedExceptions(TypeContext *ctx, TreeNode *throws);

  // widen helper functions
  virtual void _widen();
  TreeNode *typeWiden(TreeNode *from, TypeNode *to);
  TreeNode *stringWiden(TreeNode *from);
  void methodWiden();
  void methodWiden(TreeNode *args, Decl *decl);
  TypeNode *_widenSource();
  TypeNode *_castResult();

  // packageResolution helper functions
  virtual void resolvePackage (Decl* package, Environ *fileEnv);
  virtual void resolveImports(CompileUnitNode *file, Environ *fileEnv);
  virtual void resolveTypes(Decl *package, Environ *fileEnv);

  // resolveName helper functions
  virtual bool isLoop() const;

  // resolveField helper functions
  void resolveAField (bool thisAccess, bool isSuper, FieldContext *ctx);

  // resolveOperators helper functions
  virtual TreeNode *_resolveOperators(); // actually resolve operators on 'this'
  TreeNode *overload(bool assign);
  TreeNode *generalOverload(string op, bool assign,
			    TreeNode *object, TreeListNode *args);
  TreeNode *overloadArray();
  TreeNode *overloadArrayAssign();

  // constant folding helper functions
  virtual TreeNode *_fold();
  virtual Common::Kind foldMask();

  // single helper functions
  bool hasGlobalEffects(SingleState s);
  bool hasSingleEffects(SingleState s);
  SingleState duplicateState(SingleState s);
  void discardState(SingleState s);
  SingleState removeLabel(SingleState s, TreeNode *label);
  SingleState removeUnlabeledBreak(SingleState s);
  SingleState removeException(SingleState s, TypeNode *exception);
  SingleState mergeTerminations(SingleState s, TypeNode *methodType);

  void validateTerminations(SingleState s);
  SingleState sequenceState(SingleState s1, SingleState s2,
			    SingleContext *ctx);
  SingleState merge(SingleState s1, SingleState s2, bool singleSplit,
		    SingleContext *ctx);
  SingleState loopMerge(SingleState condition, SingleState body,
			SingleContext *ctx);

  static SingleState iAmSingle;
  SingleState singleList(SingleContext *ctx);
  SingleState singleSequence(SingleContext *ctx, SingleState initial = iAmSingle);
  virtual SingleState singleAssign(SingleState from, SingleContext *ctx);
  SingleState singleAssignment(SingleContext *ctx);
  SingleState singleIf(TreeNode *condition,
		       TreeNode *thenNode, TreeNode *elseNode,
		       SingleContext *ctx);
  SingleState singleCall(SingleState st,
			 bool explicitConstructorCall,
			 SingleContext *ctx);
  SingleState singleAllocateArray(SingleState st,
				  SingleContext *ctx);

#ifdef USE_SGLOBAL_INFERENCE
  // sglobal inference helper functions
public:
  virtual void sglobalData();
  virtual void sglobalCall();
protected:
  virtual void sglobalDataAssign();
#endif

protected:
  // Helper functions for dump().
  void dumpPrefix( ostream &, unsigned ) const;
  void dumpSuffix( ostream &, unsigned ) const;

  // Remember which nodes have been dumped, to sanity check replies.
  typedef set< const TreeNode *, less< const TreeNode * > > NodeSet;
  static NodeSet dumped;

  // Dataflow polyvar.
  Poly *_poly;
};

/* A TreeListNode wraps around a list of TreeNode*s to make it easy to */
/* give default definitions for the various virtual functions. */

class TreeListNode : public TreeNode 
{
public:
  TreeListNode (llist<TreeNode*>* _children, SourcePosn p = NoSourcePosition);

  const char* oper_name() const;

  void print( ostream &, unsigned ) const;
  void dump(  ostream &, unsigned ) const;

  bool compare(const TreeNode *) const; 
  int arity() const;
  bool empty() const { return !children; }

  TreeListNode *clone() const;

  void emitExpressionList( CodeContext &, vector< string > & );
  void *commonSubExpr();

  // Local qualification inference
  void localConstrainCall( const TypeListNode & ) const;
  bool localEnactDimension( const LocalEnactContext & );

protected:
  TreeListNode() { }
  size_t _arity;
};

llist<TreeNode *> *appendTreeList(TreeListNode *l1, llist<TreeNode *> *l2);

class OmittedNode : public TreeNode 
{
public:
  OmittedNode() { children = NULL; where = NoSourcePosition; }
  bool absent() const { return true; }
  void print( ostream &, unsigned ) const;
  void dump(  ostream &, unsigned ) const;
  int arity() const { return 0; }
  TypeNode* type();
  SingleState single(SingleContext *ctx);
  const string emitExpression2( CodeContext & );
  OmittedNode *clone() const { return const_cast<OmittedNode*>(this); }
};

/* An ExprNode has a type. */

class ExprNode : public TreeNode
{
  friend class ThisFieldAccessNode;
  friend class SuperFieldAccessNode;
public:
  TypeNode* type();
  void type(TypeNode *);

  /* For template arguments */
  string typeName() const; 
  bool typeIdentNM(const TreeNode*) const;

  TreeNode *instantiate(Subst *args);
  void *commonSubExpr();
  llist<CfgNode *> *makeCfg();
  const string emitExpression( CodeContext & );
  const string simpleVar( CodeContext & );
  void liftInvariants( CodeContext &, TreeNode *WRTloop );


  // As an optimization, if this string exists, it is the name of the
  // C variable that contains the value of this expression.
  string cVariable;

  // Local qualification inference
  void unshareType();
  void inferLocal( const LocalContext );
  void localEnact( const LocalEnactContext & );

protected:
  ExprNode() : _type(NULL) {}

  /* The function type above memoizes the type in the member */ 
  /* _type.  The following function does the actual computation */
  /* of the type that type() memoizes. */
  virtual TypeNode* __type() = 0;

  TypeNode* _type;

  /* Some common type checking rules */
  void bitwiseTypecheck();

  // Should never be called -- fatal error
  void codeGen( CodeContext & );

protected:
  // expression code generation helpers
  const string emitPrefixOp(    CodeContext &, const char [] );
  const string emitPostfixOp(   CodeContext &, const char [] );
  const string emitBinOp(       CodeContext &, const char [] );
  const string emitAssignBinOp( CodeContext &, const char [] );

  const string  defineTemporary( CodeContext & );
  const string declareTemporary( LocalVars & );

public:
  static const string getTemporary();
  static void resetTemporary();
  const static string declareTemporary( LocalVars &, const TypeNode & );
  const static string declareTemporary( LocalVars &, const CtType & );

private:
  static UniqueId temporary;
};


#include "cfg.h"
#include "ClassDecl.h"
#include "CtType.h"
#include "types.h"
#include "MethodDecl.h"
#include "tree-nodes.h"
#include "compiler.h"

#endif
