		/* Auxiliary lexical routines  */


#include <cstdio>
#include <iostream.h>
#include "utils.h"
#include "AST.h"
#include "tokens.h"

/* The line number on which character N of the input apppears. */
size_t File::posnToLineNumber (unsigned long N)
{
    unsigned int i;
    if (N == 0)
	return 0;
    for (i = 0; i < lineEndings.size(); i += 1)
	if (N <= lineEndings[i])
	    return i;
    return i;
}

/* Record that N is the position of a newline in the source program */
/* file.  It is assumed that this function is called with increasing */
/* values of N. */
void File::defineLineEnd (unsigned long N)
{
    lineEndings.push_back (N);
}

static int errorCount = 0;

/* Print a position in a format suitable for errors or warnings. */
ostream & operator << ( ostream &os, const SourcePosn &position )
{
    if (position.unknown())
	return os << "?: ";
    else 
	return os << *position.file->name() << ":"
		  << position.posnToLineNumber() << ": ";
}

/* Report message at source position N, giving MSG as the text. */
void Message (const string& msg, const SourcePosn N = NoSourcePosition)
{
    cerr << N << msg << endl;
}

/* Report error at source position N, giving MSG as the text of the */
/* error message. */
void Error (const string& msg, const SourcePosn N = NoSourcePosition)
{
    errorCount += 1;
    Message (msg + '.', N);
}

/* Report warning at source position N, giving MSG as the text of the */
/* error message. */
void Warning (const string& msg, SourcePosn N = NoSourcePosition)
{
    Message ("warning: " + msg + '.', N );
}

/* The number of times Error has been called. */
extern int NumErrors()
{
    return errorCount; 
}

/* An array type with given ELEMENTTYPE and DIMS dimensions.  When */
/* DIMS=0, equals ELEMENTTYPE. */
TypeNode* makeArrayType (TypeNode* elementType, int dims)
{
    while (dims > 0) {
	elementType = new JavaArrayTypeNode (elementType);
	dims -= 1;
    }
    return elementType;
}

struct InternTableEntry {
    InternTableEntry(string* datum0, InternTableEntry* next0)
	: datum(datum0), next(next0) {}
    InternTableEntry(const char* datum0, InternTableEntry* next0)
	: datum(new string(datum0)), next(next0) {}
    string* datum;
    InternTableEntry* next;
};

static InternTableEntry** internTable;
static int internTableSize = 0;

static unsigned int pjw_hash (const char* x, size_t n, size_t k)
{
    unsigned int r;
    const char* lim = x+n;

    for (r = 0; x != lim; x += 1) {
	r = (r << 4) + *x;
	if (r > 0x0fffffff) {
	    r ^= (r >> 24) & 0xf0;
	    r &= 0x0fffffff;
	}
    }
    return r % k;
}

static void createInternTable()
{
    if (internTable == NULL) {
	internTableSize = 1019;
	internTable = new InternTableEntry*[internTableSize];
	for (int i = 0; i < internTableSize; i++)
	  internTable[i] = NULL;
    }
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const char* text, size_t len)
{
    createInternTable();

    unsigned int h = pjw_hash(text, len, internTableSize);

    InternTableEntry** p;
    for (p = &internTable[h]; *p != NULL; p = &(*p)->next) {
	if (text == *(*p)->datum) 
	    return (*p)->datum;
    }
    *p = new InternTableEntry(new string(text, len), *p);
    return (*p)->datum;
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const char* text)
{
    return intern(text, strlen(text));
}

/* The unique string* containing the current contents of TEXT. */
const string* intern (const string& text)
{
    return intern(text.c_str(), text.size());
}
