#include <cassert>
#include <fstream>
#include <map>
#include "AST.h"
#include "ArraySet.h"
#include "CfHeader.h"
#include "CfSource.h"
#include "CtReference.h"
#include "CtTitaniumArray.h"
#include "CtType.h"
#include "code-assign.h"
#include "code-grid.h"
#include "code-util.h"
#include "decls.h"
#include "domain-decls.h"
#include "lgMacro.h"
#include "stl-queue.h"
#include "utils.h"


typedef map< int, string, less<int> > intstringmap;

ArraySet arrayTypes;


// For internal functions that access grids, e.g., _ti_PTR and _ti_global_READ.
string arrayInternal(const string &fn, bool global)
{
  return string("_ti_") + (global ? "global_" : "") + fn;
}

// For methods like domain, slice, etc.
string arrayMethodPrefix(const TypeNode *t)
{
  return
    (t->modifiers() & Common::Local)
    ? string("_ti_arraymethod_")
    : string("_ti_global_arraymethod_");
}

/////////////////////////////////////////////////////////////////////////////
// The section that follows handles the instantiation of grids for different
// element types and arities.
/////////////////////////////////////////////////////////////////////////////

/* Return 1 for arrays of things, 2 for arrays of arrays of things, etc. */
static int depth(const TypeNode *t)
{
  int d = 0;
  while (t->isTitaniumArrayType()) {
    d++;
    t = t->elementType();
  }
  return d;
}

static string quickconv(char *arr, int arity)
{
  string s;
  for (int i = arity; i >= 1; i--)
    s += string("(_i") + int2str(i - 1)
      + "-"
      + arr + ".base[" + int2str(i - 1) + "])/"
      + arr + ".stride[" + int2str(i - 1) + "]*"
      + arr + ".sideFactors[" + int2str(i - 1) + "]"
      + (i == 1 ? "" : "+");
  return s;
}

static string& printForallPointMacro(int arity)
{
  static intstringmap memo;
  string s;
  string& temp = memo[arity];
  int i;

  if (temp != "") return temp;

  s += "fprintf(f, \"%d";
  for (i = 1; i < arity; i++)
    s += ",%d";
  s += "\", _i0";
  for (i = 1; i < arity; i++)
    s += string(", _i") + int2str(i);
  return (memo[arity] = (s + ")"));
}

static string& forallMacro(int arity)
{
  static intstringmap memo;
  string s, is, var;
  string& temp = memo[arity];
  int i;

  if (temp != "") return temp;

  // ed is the delta to e per iteration
  is = int2str(arity - 1);
  s += "{jint ed=stride[" + is + "]/x.stride[" + is + "]*"
    "x.sideFactors[" + is + "];";

  for (i = 0; i < arity - 1; i++) {
    is = int2str(i);
    var = string("_i") + is;
    s += "{jint " + var + "=lo[" + is + "];";
    s += "for(;"
      + var + "<=hi[" + is + "];"
      + var + "+=stride[" + is + "])"
      "{"
      /* "cout<<\"" + var + "=\"<<" + var + ";" */
      ;
  }
  is = int2str(i);
  var = string("_i") + is;
  s += "{jint " + var + "=lo[" + is + "];";
  s += "PTR_TO_T e;";
  s += "INDEX(e,x.A," + quickconv("x", arity) + ");";

  s += "for(;"
    + var + "<=hi[" + is + "];"
    + var + "+=stride[" + is + "])"
    "{"
    /* "cout<<\"" + var + "=\"<<" + var + ";" */
    "body;BUMP_PTR(e,ed);}";

  for (i = 0; i < 2 * arity; i++) s += '}';

  return (memo[arity] = s);
}

static string& forall2Macro(int arity)
{
  static intstringmap memo;
  string s, is, var;
  int i;
  string& temp = memo[arity];

  if (temp != "") return temp;

  for (i = 0; i < arity - 1; i++) {
    is = int2str(i);
    var = string("_i") + is;
    s += "{jint " + var + ";";
  }
  for (i = 0; i < arity - 1; i++) {
    is = int2str(i);
    var = string("_i") + is;
    s += "for(" + var + "=lo[" + is + "];"
      + var + "<=hi[" + is + "];"
      + var + "+=stride[" + is + "]) ";
  }
  is = int2str(i);
  var = string("_i") + is;
  // ed is the delta to e per iteration, fd is the delta to f
  s += "{jint fd,ed," + var + "=lo[" + is + "];";
  s += "PTR_TO_T e,f;";
  s += "INDEX(e,x.A," + quickconv("x", arity) + ");";
  s += "INDEX(f,y.A," + quickconv("y", arity) + ");";
  s+= "ed=stride[" + is + "]/x.stride[" + is + "]*x.sideFactors[" + is + "];"
    + "fd=stride[" + is + "]/y.stride[" + is + "]*y.sideFactors[" + is + "];";
  s += "for(;"
    + var + "<=hi[" + is + "];"
    + var + "+=stride[" + is + "])"
    "{body;BUMP_PTR(e,ed);BUMP_PTR(f,fd);}";

  for (i = 0; i < arity; i++) s += '}';

  return (memo[arity] = s);
}

#define D(name, val, stmt)				\
    f << "#define " << name << ' ' << val << '\n';	\
    stmt;						\
    f << "#undef " << name << '\n';

#define E(name, args, val, stmt)					\
    f << "#define " << name << "(" << args << ") " << val << '\n';	\
    stmt;								\
    f << "#undef " << name << '\n';

#define ARRAY_TYPE(t, n, global) \
    (((global) ? "TI_GLOBAL_ARR_" : "TI_ARR_") + (t) + '_' + int2str(n))
static void instantiateToFile(const TypeNode &type, CfCode &f)
{
  const CtTitaniumArray &cType = static_cast<const CtTitaniumArray &>(type.cType());
  const string t = type.elementType()->cType().name;
  const bool global = !type.isLocal();
  const int n = type.tiArity();
  const char *s = global ? "_ti_global_" : "_ti_";
  const AssignKind arrayKind = assignKind(type.elementType());
  const bool isAtomic = type.elementType()->isAtomic();

  const CtType *arraySliceType;
  if (n > 1) {
    LitNode *arity = new LitNode( (int32) (n - 1) );
    TypeNode *elem = type.elementType();
    TypeNode *sliced = new TitaniumArrayTypeNode( elem, arity );
    if (type.isLocal()) sliced = sliced->addModifiers( Common::Local );	
    arraySliceType = &sliced->cType();
  } else
    arraySliceType = &type.cType(); // arbitrary

  type.cType().define(f);
  type.elementType()->cType().define(f);
  arraySliceType->define(f);
  RectDomainNDecl[ n - 1 ]->includeSelf(f);

  D("ti_ARRAY", cType,
  D("ti_ARRAY_SLICE", *arraySliceType,
  D("ti_RECTDOMAIN", "_ti_RECTDOMAIN(" << n << ")",
  D("ti_DOMAIN", "_ti_DOMAIN(" << n << ")",
  D("ti_POINT", "P" << n,
  D("ti_conv", s << "CONV(" << t << ", " << n << ")",
  D("ti_dump", s << "DUMP(" << t << ", " << n << ")",
  D("ti_read", s << "READ(" << t << ", " << n << ")",
  D("ti_ptr", s << "PTR(" << t << ", " << n << ")",
  D("ti_write", s << "WRITE(" << t << ", " << n << ")",
  D("ti_construct", s << "CONSTRUCT(" << t << ", " << n << ")",
  D("ti_empty", s << "EMPTY(" << t << ", " << n << ")",
  D("ti_isnull", s << "ISNULL(" << t << ", " << n << ")",
  D("DOMAIN_METHOD", s << "arraymethod_domain(" << t << ", " << n << ")",
  D("BROADCAST_METHOD", s << "arraymethod_broadcast(" << t << ", " << n << ")",
  D("COPY_METHOD", s << "arraymethod_copy(" << t << ", " << n << ")",
  D("ISLOCAL_METHOD", s << "arraymethod_isLocal(" << t << ", " << n << ")",
  D("CREATOR_METHOD", s << "arraymethod_creator(" << t << ", " << n << ")",
  D("RESTRICT_METHOD", s << "arraymethod_restrict(" << t << ", " << n << ")",
  D("EXCHANGE_METHOD", s << "arraymethod_exchange(" << t << ", " << n << ")",
  D("PERMUTE_METHOD", s << "arraymethod_permute(" << t << ", " << n << ")",
  D("SET_METHOD", s << "arraymethod_set(" << t << ", " << n << ")",
  D("SLICE_METHOD", s << "arraymethod_slice(" << t << ", " << n << ")",
  D("TRANSLATE_METHOD", s << "arraymethod_translate(" << t << ", " << n << ")",
  D("INJECT_METHOD", s << "arraymethod_inject(" << t << ", " << n << ")",
  D("PROJECT_METHOD", s << "arraymethod_project(" << t << ", " << n << ")",
  D("POINT_INDEX", "_POINT_INDEX(" << n << ")",
  D("POINT_SET", "_POINT_SET(" << n << ")",
  E("RECTDOMAIN_EMPTY", "",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_tiRectDomain" << n << "())",
  E("RECTDOMAIN_SLICE", "R, k",
	((n == 3) ? "slice32((R), (k))" :
	 ((n == 2) ? "slice21((R), (k))" : "assert(0)")),
  E("RECTDOMAIN_INTERSECTION", "a, b",
    "(cl_ti_domains_tiRectDomain" << n <<
    "_mt_multiply_cl_ti_domains_tiRectDomain" << n << "((a), (b)))",
  E("RECTDOMAIN_STRIDE", "a",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_getLoopingStride(a))",
  E("RECTDOMAIN_MIN", "a",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_getLowerBound(a))",
  E("RECTDOMAIN_MAX", "a",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_getUpperMaxPoint(a))",
  E("RECTDOMAIN_UPB", "a",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_getUpperBound(a))",
  E("RECTDOMAIN_ISNULL", "a",
    "(cl_ti_domains_tiRectDomain" << n << "_mt_isNull(a))",

  E("RECTDOMAIN_TRANSLATE", "a, b",
    "(cl_ti_domains_tiRectDomain" << n <<
    "_mt_translate_cl_ti_domains_tiPoint" << n << "((a), (b)))",
  E("RECTDOMAIN_PERMUTE", "a, b",
    "(cl_ti_domains_tiRectDomain" << n <<
    "_mt_permute_cl_ti_domains_tiPoint" << n << "((a), (b)))",
  E("RECTDOMAIN_INJECT", "a, b",
    "(cl_ti_domains_tiRectDomain" << n <<
    "_mt_inject_cl_ti_domains_tiPoint" << n << "((a), (b)))",
  E("RECTDOMAIN_PROJECT", "a, b",
    "(cl_ti_domains_tiRectDomain" << n <<
    "_mt_project_cl_ti_domains_tiPoint" << n << "((a), (b)))",

  D("N_MINUS_1", (n - 1),
  D("T_string", ('"' + t + '"'),
  D("T", t,
  D("ELEMENTS_ARE_ATOMIC", (isAtomic ? 1 : 0),
  E("BUMP_PTR", "p, delta", "INDEX(p, p, delta)",
  D("GLOBAL_ARRAY", (global ? 1 : 0),
  D("GLOBAL_ELEMENTS", (type.elementType()->isLocal() ? 0 : 1),
  D("assign", assignMacro(arrayKind, false, !global),
  D("weak_assign", "WEAK_" << assignMacro(arrayKind, false, !global),
  D("N", n,
  E("forall", "e, x, lo, hi, stride, body", forallMacro(n),
  E("forall2", "e, x, f, y, lo, hi, stride, body", forall2Macro(n),
  E("PRINT_FORALL_POINT", "f", printForallPointMacro(n),
  D("PTR_TO_T", cType.data,

  f << "#define TI_ARRAY_ADDR_"
    << (global ? "GLOBAL" : "LOCAL")
    << '_' << t << '_' << n
    << "(result, array, index)  ((result) = "
    << arrayInternal("PTR", global)
    << '(' << t << ", " << n << ')'
    << "(\"ti_array_addr\", (array), (index)))\n"

    << "#include <ti_array." << f.suffix << ">\n"
    ))))))))))))))))))))))))))))))))))))))))))))))))))))));
}

static void instantiateArrayCasts( const TypeNode &local,
				   const TypeNode &global,
				   CfCode &f, bool header )
{
  const CtType &t = local.elementType()->cType();
  const string n = int2str(local.tiArity());

  if (header) {
    global.cType().declare(f);
    local.cType().declare(f);
  } else {
    global.cType().define(f);
    local.cType().define(f);
  }

  D("ti_GLOBAL_ARRAY", global.cType(),
  D("ti_LOCAL_ARRAY", local.cType(),
  D("ti_cast_to_global", "_ti_arraymethod_cast(" << t << ", " << n << ")",
  D("ti_cast_to_local",
     "_ti_global_arraymethod_cast(" << t << ", " << n << ")",

  if (header)
    f <<
      "/* Functions defined in ti_cast_array.c */\n"
      "ti_GLOBAL_ARRAY ti_cast_to_global(const ti_LOCAL_ARRAY *l, int proc);\n"
      "ti_LOCAL_ARRAY ti_cast_to_local(const char *where, const ti_GLOBAL_ARRAY *g);\n";
  else
    f << "#include <ti_cast_array.c>\n";

  ))));
}


static void instantiatePair( const TypeNode &local,
			     const TypeNode &global,
			     CfCode &code, bool header )
{
  instantiateToFile(  local, code );
  instantiateToFile( global, code );
  instantiateArrayCasts( local, global, code, header );
}


static void instantiateArrays( const TitaniumArrayTypeNode &type )
{
  if (DEBUG)
    cout << "  adding array <" << type.errorName() << ">\n";

  const TypeNode &local  = *type.addModifiers(Common::Local);
  const TypeNode &global = *type.removeModifiers(Common::Local);

  CfHeader header( global.cType() );
  CfSource source( global.cType() );

  // Create local and global versions of all arrays, whether we need
  // them or not.  This is slightly wasteful.
  instantiatePair( local, global, header, true );
  instantiatePair( local, global, source, false );
}


void instantiateArrays(void)
{
  for (ArraySet::const_iterator sweep = arrayTypes.begin();
       sweep != arrayTypes.end(); ++sweep)
    instantiateArrays(**sweep);
}
