// Java literal class, with associated operations. This code assumes that
// the rules for integer and floating point arithmetic are the same as
// for java (i.e. 2's complement integer arithmetic, IEEE floating point
// with the correct options set)

#include <cassert>
#include <cctype>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iomanip.h>
#include <strstream.h>
#include "AST.h"
#include "runtime/fp-utils.h"


int64 Literal::intValue() const
{
  switch (kind())
    {
    case Common::ByteKind: return (int8)i;
    case Common::ShortKind: return (int16)i;
    case Common::CharKind: return (uint16)i;
    case Common::IntKind: return (int32)i;
    case Common::LongKind: return i;
    default: assert(0);
    }
}

Literal Literal::arithPromote()
{
  switch (kind())
    {
    case Common::ByteKind: case Common::ShortKind: case Common::CharKind:
      return cast(Common::IntKind);
    default: return *this;
    }
}

Literal Literal::arithPromote(Common::Kind to)
{
  if (kind() == Common::DoubleKind || to == Common::DoubleKind)
    return cast(Common::DoubleKind);
  if (kind() == Common::FloatKind || to == Common::FloatKind)
    return cast(Common::FloatKind);
  if (kind() == Common::LongKind || to == Common::LongKind)
    return cast(Common::LongKind);
  if (kind() == Common::BoolKind || to == Common::BoolKind)
    return cast(Common::BoolKind);

  return cast(Common::IntKind);
}

Literal Literal::cast(Common::Kind to)
{
  Literal me = *this;

  switch (kind())
    {
    case Common::BoolKind:
      assert(to == Common::BoolKind);
      break;
    case IntegerKind:
      switch (to)
	{
	case Common::ByteKind: me.i = (int8)i; break;
	case Common::ShortKind: me.i = (int16)i; break;
	case Common::CharKind: me.i = (uint16)i; break;
	case Common::IntKind: me.i = (int32)i; break;
	case Common::LongKind: me.i = (int64)i; break;
	case Common::FloatKind: me.f = i; break;
	case Common::DoubleKind: me.d = i; break;
	case Common::BoolKind: assert(0);
	default: assert(0);
	}
      break;
    case Common::FloatKind:
      switch (to)
	{
	case Common::ByteKind: me.i = (int8)f; break;
	case Common::ShortKind: me.i = (int16)f; break;
	case Common::CharKind: me.i = (uint16)f; break;
	case Common::IntKind: me.i = (int32)f; break;
	case Common::LongKind: me.i = (int64)f; break;
	case Common::FloatKind: break;
	case Common::DoubleKind: me.d = f; break;
	case Common::BoolKind: assert(0);
	default: assert(0);
	}
      break;
    case Common::DoubleKind:
      switch (to)
	{
	case Common::ByteKind: me.i = (int8)d; break;
	case Common::ShortKind: me.i = (int16)d; break;
	case Common::CharKind: me.i = (uint16)d; break;
	case Common::IntKind: me.i = (int32)d; break;
	case Common::LongKind: me.i = (int64)d; break;
	case Common::FloatKind: me.f = d;
	case Common::DoubleKind: break;
	case Common::BoolKind: assert(0);
	default: assert(0);
	}
      break;
    default: assert(0);
    }
  me._kind = to;

  return me;
}

Literal Literal::operator~()
{
  Literal me = arithPromote();

  switch (me.kind())
    {
    case IntegerKind: me.i = ~me.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator!()
{
  assert(kind() == Common::BoolKind);

  return Literal((bool)!b);
}

Literal Literal::operator-()
{
  Literal me = arithPromote();

  switch (me.kind())
    {
    case IntegerKind: me.i = -me.intValue(); break;
    case Common::FloatKind: me.f = -me.f; break;
    case Common::DoubleKind: me.d = -me.doubleValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator*(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() * arg.intValue(); break;
    case Common::FloatKind: me.f = me.f * arg.f; break;
    case Common::DoubleKind: me.d = me.doubleValue() * arg.doubleValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator/(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() / arg.intValue(); break;
    case Common::FloatKind: me.f = me.f / arg.f; break;
    case Common::DoubleKind: me.d = me.doubleValue() / arg.doubleValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator%(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() % arg.intValue(); break;
    case Common::FloatKind: me.f = fmod(me.f, arg.f); break;
    case Common::DoubleKind: me.d = fmod(me.doubleValue(), arg.doubleValue()); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator+(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() + arg.intValue(); break;
    case Common::FloatKind: me.f = me.f + arg.f; break;
    case Common::DoubleKind: me.d = me.doubleValue() + arg.doubleValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator-(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() - arg.intValue(); break;
    case Common::FloatKind: me.f = me.f - arg.f; break;
    case Common::DoubleKind: me.d = me.doubleValue() - arg.doubleValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::lsl(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = me.intValue() << arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::rsl(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.i = (unsigned long long)me.intValue() >> arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::rsa(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
      // warning: this assumes that >> on signed quantities is arithmetic
      // right shift (not guaranteed by ANSI C, at least)
    case IntegerKind: me.i = me.intValue() >> arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::lt(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.b = me.intValue() < arg.intValue(); break;
    case Common::FloatKind: me.b = me.f < arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() < arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::le(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.b = me.intValue() <= arg.intValue(); break;
    case Common::FloatKind: me.b = me.f <= arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() <= arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::gt(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.b = me.intValue() > arg.intValue(); break;
    case Common::FloatKind: me.b = me.f > arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() > arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::ge(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case IntegerKind: me.b = me.intValue() >= arg.intValue(); break;
    case Common::FloatKind: me.b = me.f >= arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() >= arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::eq(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  if (me.kind() != arg.kind()) return false;

  switch (me.kind())
    {
    case Common::BoolKind: me.b = me.boolValue() == arg.boolValue(); break;
    case IntegerKind: me.b = me.intValue() == arg.intValue(); break;
    case Common::FloatKind: me.b = me.f == arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() == arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::ne(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  if (me.kind() != arg.kind()) return true;

  switch (me.kind())
    {
    case Common::BoolKind: me.b = me.boolValue() != arg.boolValue(); break;
    case IntegerKind: me.b = me.intValue() != arg.intValue(); break;
    case Common::FloatKind: me.b = me.f != arg.f; break;
    case Common::DoubleKind: me.b = me.doubleValue() != arg.doubleValue(); break;
    default: assert(0);
    }

  me._kind = Common::BoolKind;
  return me;
}

Literal Literal::operator&(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case Common::BoolKind: me.b = me.boolValue() && arg.boolValue(); break;
    case IntegerKind: me.i = me.intValue() & arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator|(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case Common::BoolKind: me.b = me.boolValue() || arg.boolValue(); break;
    case IntegerKind: me.i = me.intValue() | arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator^(Literal arg)
{
  Literal me = arithPromote(arg.kind());
  arg = arg.arithPromote(me.kind());

  switch (me.kind())
    {
    case Common::BoolKind: me.b = me.boolValue() ^ arg.boolValue(); break;
    case IntegerKind: me.i = me.intValue() ^ arg.intValue(); break;
    default: assert(0);
    }

  return me;
}

Literal Literal::operator&&(Literal arg)
{
  Literal me = *this;

  assert(me.kind() == Common::BoolKind && arg.kind() == Common::BoolKind);
  me.b = me.boolValue() && arg.boolValue();

  return me;
}

Literal Literal::operator||(Literal arg)
{
  Literal me = *this;

  assert(me.kind() == Common::BoolKind && arg.kind() == Common::BoolKind);
  me.b = me.boolValue() || arg.boolValue();

  return me;
}

ostream& operator<<(ostream& os, Literal l)
{
  switch (l.kind())
    {
    case Common::ByteKind: os << "jbyte(" << l.intValue() << ")"; break;
    case Common::ShortKind: os << "jshort(" << l.intValue() << ")"; break;
    case Common::CharKind: os << "jchar(" << l.intValue() << ")"; break;
    case Common::IntKind: os << "jint(" << l.intValue() << ")"; break;
    case Common::LongKind: os << "jlong(" << l.intValue() << ")"; break;
    case Common::FloatKind: os << "jfloat(" << l.doubleValue() << ")"; break;
    case Common::DoubleKind: os << "jdouble(" << l.doubleValue() << ")"; break;
    case Common::BoolKind: os << "jboolean(" << l.boolValue() << ")"; break;
    default: assert(0);
    }

  return os;
}

Literal intLiteral(const string &s)
{
  int32 value;
  istrstream source( s.c_str() );
  source >> setbase( 0 ) >> value;
  return Literal( value );
}

Literal longLiteral(const string &s)
{
  int64 value;
  istrstream source( s.c_str() );
  source >> setbase( 0 ) >> value;
  return Literal( value );
}

Literal charLiteral(const string &s)
{
  if (s.size() > 1) // must be a \uxxxx sequence
    return Literal((uint16)strtol(s.c_str() + 2, NULL, 16));
  else
    return Literal((uint16)s[0]);
}

Literal floatLiteral(const string &s)
{
  float value = strtod( s.c_str(), 0 );
  return Literal( value );
}

Literal doubleLiteral(const string &s)
{
  double value = strtod( s.c_str(), 0 );
  return Literal( value );
}

const string Literal::asString() const
{
  ostrstream format;
  format.setf( ios::hex | ios::showbase,
	       ios::basefield | ios::showbase );
  
  switch (kind())
    {
    case Common::BoolKind:
      return boolValue() ? "1" : "0";
    case Common::ByteKind:
      format << (int8) intValue();
      break;
    case Common::CharKind:
      format << (uint16) intValue();
      break;
    case Common::ShortKind:
      format << (int16) intValue();
      break;
    case Common::IntKind:
      format << (int32) intValue();
      break;
    case Common::LongKind:
      format << (int64) intValue() << "LL";
      break;
    case Common::FloatKind:
    case Common::DoubleKind: {
      char buffer[ 32 ];
      g_fmt( buffer, doubleValue() );      
      string result = buffer;
      if (result.find_first_of(".eE") == string::npos)
	result += '.';      
      return result;
    }
    default: assert(0);
    }

  string result( format.str(), format.pcount() );
  format.freeze( 0 );
  
  return result;
}
