/* static.cc: Main module for Java static semantics. */

#include <fstream.h>
#include "AST.h"
#include "FieldDecl.h"
#include "MethodDecl.h"
#include "NameContext.h"
#include "code-point.h"
#include "code-util.h"
#include "decls.h"
#include "domain-decls.h"
#include "is-main.h"
#include "lexer.h"
#include "utils.h"
#include "errno.h"
#include "errors.h"
#include "local/options.h"
#include "local/traverse.h"
#include "st-sglobal.h"

const string
  *TiString,
  *ObjectString,
  *UnknownString,
  *StringString,
  *StringBufferString,
  *ErrorString,
  *RuntimeExceptionString,
  *ClassCastExceptionString,
  *ThrowableString,
  *JavaArrayString,
  *RectDomainM1String,
  *RectDomainString,
  *DomainString,
#if 1 // !!! HACK miyamoto: to always instantiate domains
  *PointNString[3],
  *RectDomainNString[3],
  *RectDomainListNString[3],
  *RectDomainListIteratorNString[3],
  *DomainNString[3],
#endif
  *TiArrayString,
  *TiArrayLString,
  *TiArrayM1String,
  *TemplateArgString,
  *PointString,
  *CloneableString,
  *TitaniumString,
  *InternalString,
#if 1 // !!! HACK miyamoto: to always instantiate domains
  *DomainLibString,
#endif
  *JavaString,
  *LangString,
  *ArityString,
  *RegionString,
  *SharedRegionString,
  *PrivateRegionString;

Decl
  *unnamedPackage;

ClassDecl
  *ObjectDecl,
  *StringDecl,
  *StringBufferDecl,
  *ErrorDecl,
  *RuntimeExceptionDecl,
  *ClassCastExceptionDecl,
  *ThrowableDecl,
  *JavaArrayDecl,
  *RectDomainM1Decl,
  *RectDomainDecl,
  *DomainDecl,
  *PointDecl,
#if 1 // !!! HACK miyamoto: to always instantiate domains
  *PointNDecl[3],
  *DomainNDecl[3],
  *RectDomainNDecl[3],
  *RectDomainListNDecl[3],
  *RectDomainListIteratorNDecl[3],
#endif
  *TiArrayM1Decl,
  *TiArrayDecl,
  *TiArrayLDecl,
  *TemplateArgDecl,
  *CloneableDecl,
  *RegionDecl,
  *SharedRegionDecl,
  *PrivateRegionDecl,
  *NativeUtilsDecl;

Decl
  *TiArityDecl;

ClassDecl
  *UnknownClassDecl;

Decl
  *UnknownPackageDecl,
  *UnknownFieldDecl,
  *UnknownMethodDecl;

TypeNode
  *theArrayInitializerType;

extern bool opt_dataflow;

static void importPackage(Environ *env, const string *s1, const string *s2)
{
  TreeNode* name =
    new NameNode(new NameNode(TreeNode::omitted, s1, NULL), s2, NULL);

  int errs0 = NumErrors();
  resolveAName(name, *(PackageDecl::System->environ()), NULL, false, NULL,
	       Decl::Package);

  if (errs0 == NumErrors())
    foriter (type, name->decl()->environ()->allProperDecls(), EnvironIter)
      if (type->category() != Decl::Package)
	env->add(&*type); // conflicts appear on use only
}

static ClassDecl *requireClass(const Environ &env, const string *name)
{
  ClassDecl * const decl = static_cast< ClassDecl * >(env.lookup( name ));
  assert( decl->category() & (Decl::Class | Decl::Interface) );
  decl->drequire();
  return decl;
}


/* Initialize Decls, types, etc, used internally */
void initTheSystem()
{
  theArrayInitializerType = new ArrayInitializerTypeNode;
  
  TiString = intern ("Ti");
  ObjectString = intern ("Object");
  UnknownString = intern("tiUnknown");
  StringString = intern ("String");
  StringBufferString = intern ("StringBuffer");
  ErrorString = intern ("Error");
  RuntimeExceptionString = intern ("RuntimeException");
  ClassCastExceptionString = intern ("ClassCastException");
  ThrowableString = intern ("Throwable");
  CloneableString = intern ("Cloneable");
  JavaArrayString = intern ("tiJArray");
  DomainString = intern ("tiDomain");
  RectDomainString = intern ("tiRectDomain");
  RectDomainM1String = intern ("tiRectDomainM1");
  PointString = intern ("tiPoint");
  RegionString = intern("Region");
  SharedRegionString = intern("SharedRegion");
  PrivateRegionString = intern("PrivateRegion");

#if 1 // !!! HACK miyamoto: to always instantiate domains
  PointNString[0] = intern ("tiPoint1");
  DomainNString[0] = intern ("tiDomain1");
  RectDomainNString[0] = intern ("tiRectDomain1");
  RectDomainListNString[0] = intern ("tiRectDomainList1");
  RectDomainListIteratorNString[0] = intern ("tiRectDomainListIterator1");

  PointNString[1] = intern ("tiPoint2");
  DomainNString[1] = intern ("tiDomain2");
  RectDomainNString[1] = intern ("tiRectDomain2");
  RectDomainListNString[1] = intern ("tiRectDomainList2");
  RectDomainListIteratorNString[1] = intern ("tiRectDomainListIterator2");

  PointNString[2] = intern ("tiPoint3");
  DomainNString[2] = intern ("tiDomain3");
  RectDomainNString[2] = intern ("tiRectDomain3");
  RectDomainListNString[2] = intern ("tiRectDomainList3");
  RectDomainListIteratorNString[2] = intern ("tiRectDomainListIterator3");
#endif
  TiArrayString = intern("tiArray");
  TiArrayLString = intern("tiArrayL");
  TiArrayM1String = intern("tiArrayM1");
  TemplateArgString = intern("tiTemplateArgument");
  JavaString = intern ("java");
  TitaniumString = intern ("ti");
  InternalString = intern ("internal");
#if 1 // !!! HACK miyamoto: to always instantiate domains
  DomainLibString = intern ("domains");
#endif
  LangString = intern ("lang");
  ArityString = intern ("__arity");

  unnamedPackage = new PackageDecl(new string(""));

  Environ env;
  importPackage(&env, JavaString, LangString);
  importPackage(&env, TitaniumString, LangString);
  importPackage(&env, TitaniumString, InternalString);
  importPackage(&env, TitaniumString, DomainLibString);

  ObjectDecl = requireClass( env, ObjectString );
  UnknownClassDecl = requireClass( env, UnknownString );
  StringDecl = requireClass( env, StringString );
  StringBufferDecl = requireClass( env, StringBufferString );
  ErrorDecl = requireClass( env, ErrorString );
  RuntimeExceptionDecl = requireClass( env, RuntimeExceptionString );
  ClassCastExceptionDecl = requireClass( env, ClassCastExceptionString );
  ThrowableDecl = requireClass( env, ThrowableString );
  CloneableDecl = requireClass( env, CloneableString );
  JavaArrayDecl = requireClass( env, JavaArrayString );
  RegionDecl = requireClass( env, RegionString );
  SharedRegionDecl = requireClass( env, SharedRegionString );
  PrivateRegionDecl = requireClass( env, PrivateRegionString );
  
#ifndef JAVA
  TiArrayDecl = requireClass( env, TiArrayString );
  TiArrayLDecl = requireClass( env, TiArrayLString );
  TiArrayM1Decl = requireClass( env, TiArrayM1String );
  TemplateArgDecl = requireClass( env, TemplateArgString );
  PointDecl = requireClass( env, PointString );
  DomainDecl = requireClass( env, DomainString );
  RectDomainDecl = requireClass( env, RectDomainString );
  RectDomainM1Decl = requireClass( env, RectDomainM1String );
#endif

#if 1 // !!! HACK miyamoto: to always instantiate domains
  for (int i = 0 ; i < 3; i++) {
    PointNDecl[i] = requireClass( env, PointNString[i] );
    DomainNDecl[i] = requireClass( env, DomainNString[i] );
    RectDomainNDecl[i] = requireClass( env, RectDomainNString[i] );
    RectDomainListNDecl[i] = requireClass( env, RectDomainListNString[i] );
    RectDomainListIteratorNDecl[i] = requireClass( env, RectDomainListIteratorNString[i] );
  }
#endif

  /* Instantiate java.lang.UNIXProcess and related stuff */
  requireClass( env, intern( "UNIXProcess" ) );
  requireClass( env, intern( "OutOfMemoryError" ) );

  requireClass( env, TiString );
  NativeUtilsDecl = requireClass( env, intern( "NativeUtils" ) );
  PointUtilsDecl = requireClass( env, intern( "PointUtils" ) );

#ifdef USE_LOCAL_INFERENCE
  if (opt_local)
    requireClass( env, intern("LocalInfer") );
#endif // USE_LOCAL_INFERENCE

  // string ti_lang_ti = new string("ti/lang/Ti.java");
  // load(ti_lang_ti, ClassLib.fopen(ti_lang_ti, "r"));

  UnknownPackageDecl = new PackageDecl (NULL);
  UnknownPackageDecl->environ (new Environ);
  // Unkown field & method are public & static to avoid cascading errors
  UnknownFieldDecl = new FieldDecl (NULL,
				    theIntType, UnknownClassDecl,
				    (Common::Modifiers) 
				    (Common::Static | Common::Public),
				    TreeNode::omitted);
  UnknownMethodDecl =
    new MethodDecl (NULL,
		    new MethodTypeNode (NULL, theIntType, NULL),
		    Decl::Method,
		    UnknownClassDecl, 
		    (Common::Modifiers)(Common::Static | Common::Public),
		    TreeNode::omitted);

  /* build environments for runtime stuff and find tiArity field */
  buildEnvironments();
#ifndef JAVA
  TiArityDecl = TemplateArgDecl->environ()->lookup(ArityString);
#endif
}

llist<TreeNode *> *allFiles;
static llist<TreeNode *> *recentFiles, *unresolvedFiles;
static TreeNode *loaded;

/* The parser's interface to the rest of the compiler.  The */
/* root of the synthesized abstract syntax tree is passed to */
/* this function. */
void compileAST (TreeNode* root)
{
  root->packageResolution();
  loaded = root;
}

bool isTitanium(string fileName)
{
  const string::size_type length = fileName.size();
  return length > 3 && fileName.substr(length - 3) == ".ti";
}

void load(const string name, bool titaniumFile)
{
  load(name.c_str(), titaniumFile);
}

void load(const char * const name, bool titaniumFile)
{
  FILE * const file = fopen(name, "r");
  load(name, file, titaniumFile);
  if (file) fclose(file);
}

void load(string name, FILE *file, bool titaniumFile)
{
  if (file) {
    yyin = file;
    loaded = NULL;
    parse(name, titaniumFile);
    if (loaded)
      {
	loaded->ident(new string(name));
	allFiles = cons(loaded, allFiles);
	recentFiles = cons(loaded, recentFiles);
	unresolvedFiles = cons(loaded, unresolvedFiles);
      }
  } else
    Error("Couldn't open " + name + ":" + strerror(errno));
}

void buildEnvironments()
{
  foreach (f, llist<TreeNode *>, *recentFiles)
    (*f)->resolveClass(NULL, NULL, NULL);

  foreach (f, llist<TreeNode *>, *recentFiles)
    (*f)->resolveInheritance();

  free_all(recentFiles);
  recentFiles = NULL;
}

void setupBblocksMethod(TreeNode *t);
void setupUdMethod(MethodDeclNode *t);
void polyPrint(TreeNode *t);
void fwdSub(TreeNode *t);

// Calls "action" on all the methods found in "file".
static void processMethods(TreeNode *file, void action(TreeNode *t)) {
  const int childCount = file->arity();

  if (isMethodDeclNode(file)) {
    action(file);
  }
  else {
    for (int sweep = 0; sweep < childCount; sweep++) {
      // Need to quit the recursion when it's obvious there isn't
      // going to be a MethodDeclNode here.
      processMethods(file->child(sweep), action);
    }
  }
}

// Static analysis on one method.
static void staticAnalyze(TreeNode *method) {
  setupBblocksMethod(method);
  setupUdMethod((MethodDeclNode *)method);
}

void staticSemantics()
{
  buildEnvironments();

  extern bool canbuildenv;
  canbuildenv = true;

  while (unresolvedFiles)
    {
      TreeNode *first = unresolvedFiles->front();
      unresolvedFiles = unresolvedFiles->free();

      if (DEBUG)
	cout << "resolveName on " << *first->ident() << "\n";
      first->resolveName(NULL);
    }

  extern bool allshouldbeloaded;
  allshouldbeloaded = true;	// catch bugs (no more loads after this point,
				// see src-input.cc)

  // Handle files in the order they were loaded
  allFiles = dreverse(allFiles);

  verifyCircularity();

  foldConstants();

  foreach (f, llist<TreeNode *>, *allFiles)
    {
      TreeNode *file = *f;

      if (DEBUG)
	cout << "static resolution of " << *file->ident() << "\n";
      file->resolveField(NULL, NULL);
      file->resolveConcat();
      file->typecheck(NULL);
    }

  if (NumErrors() == 0)
    {
      if (use_sglobal_inference)
	sglobalInference();
	
      foreach (f, llist<TreeNode *>, *allFiles)
	singleAnalysis(*f);
    }

  if (NumErrors() == 0)
    foreach (f, llist<TreeNode *>, *allFiles)
      (*f)->rewrite( 0 );

  if (NumErrors() == 0)
    {
      templateEnv.resolveTemplates();
      inferLocal();
  
      foreach (f, llist<TreeNode *>, *allFiles)
	{
	  TreeNode &file = **f;
	  if (file.selectedForCodeGen(0))
	    {
	      file.widen();
	  
	      if (opt_dataflow) {
		cout << "Analyzing: ";
		cout << file.ident()->c_str() << endl;
		llist<TreeNode *> *t1 = NULL, *t2 = NULL;
		file.introduceTemporary(NULL, t1, t2);
		file.makeCfg();
		// processMethods(&file, staticAnalyze);
	      }

	      file.resolveRequires(0);
	    }
	  file.lookForMain();
	}
    }
  checkOnlyOneMain();
}

				/* The decl method. */

ClassDecl* TypeNameNode::decl() const
{
  Decl * const result = name()->decl();
  assert( !result || result->category() & (Decl::Class | Decl::Interface) );
  return static_cast< ClassDecl * >( result );
}

TypeDecl *JavaArrayTypeNode::decl() const
{
  return getJavaArrayDecl(this);
}

ClassDecl *PointTypeNode::decl() const
{
  return getPointDecl(this);
}

ClassDecl *DomainTypeNode::decl() const
{
  return getDomainDecl(this);
}

ClassDecl *RectDomainTypeNode::decl() const
{
  return getRectDomainDecl(this);
}

ClassDecl *TitaniumArrayTypeNode::decl() const
{
  return getTiArrayDecl(this);
}

Decl* ObjectNode::decl() const
{
  return ObjectNode::name()->decl();
}

ClassDecl* ClassDeclNode::decl() const
{
  Decl * const result = simpName()->decl();
  assert( result->category() == Decl::Class );
  return static_cast< ClassDecl * >( result );
}

Decl* FieldDeclNode::decl() const
{
  return simpName()->decl();
}

Decl* VarDeclNode::decl() const
{
  return simpName()->decl();
}

MethodDecl* MethodDeclNode::decl() const
{
  Decl * const result = simpName()->decl();
  assert( result->category() == Decl::Method );
  return static_cast< MethodDecl * >(result);
}

MethodDecl* ConstructorDeclNode::decl() const
{
  Decl * const result = simpName()->decl();
  assert( result->category() == Decl::Constructor );
  return static_cast< MethodDecl * >(result);
}

Decl* MethodSignatureNode::decl() const
{
  return simpName()->decl();
}

ClassDecl* InterfaceDeclNode::decl() const
{
  Decl * const result = simpName()->decl();
  assert( result->category() == Decl::Interface );
  return static_cast< ClassDecl * >( result );
}

Decl* ParameterNode::decl() const
{
  return simpName()->decl();
}

Decl* FieldAccessNode::decl() const
{
  return simpName()->decl();
}

Decl* MethodCallNode::decl() const
{
  return method()->decl();
}

Decl* MethodCallAssignNode::decl() const
{
  return method()->decl();
}

TreeNode* NameNode::simpName() const
{
  return const_cast<TreeNode*> ((const TreeNode*) this);
}
